/*---------------------------------------------------------------
 *	PWMを使ったドレミファ音階の再生(μT-Kernel 3.0用)
 *
 *	Copyright (C) 2022-2023 by T3 WG of TRON Forum
 *---------------------------------------------------------------*/
#include <tk/tkernel.h>
#include <tm/tmonitor.h>

//-------- PWMのレジスタ定義(※A) -------------------------------
#define PWM_BASE(un)	((un) == 0 ? 0x4001c000 : (un) == 1 ? 0x40021000 : (un) == 2 ? 0x40022000 : 0x4002d000)
#define PWM(un, r)				(PWM_BASE(un) + PWM_##r)
#define PWM_TASKS_STOP			0x004
#define PWM_TASKS_SEQSTART(n)	(0x008 + (n) * 0x04)
#define PWM_LOOPSDONE			0x11C
#define PWM_ENABLE				0x500
#define PWM_MODE				0x504
#define PWM_COUNTERTOP			0x508
#define PWM_PRESCALER			0x50c
#define PWM_DECODER				0x510
#define PWM_LOOP				0x514
#define PWM_SEQ_PTR(n)			(0x520 + (n) * 0x20)
#define PWM_SEQ_CNT(n)			(0x524 + (n) * 0x20)
#define PWM_SEQ_REFRESH(n)		(0x528 + (n) * 0x20)
#define PWM_SEQ_ENDDELAY(n)		(0x52c + (n) * 0x20)
#define PWM_PSEL_OUT(n)			(0x560 + (n) * 0x04)

// 比較値COMP0を格納する16ビット幅のメモリ領域
static _UH pwm_comp0;

// unitのPWMに周波数freq(Hz)を出力するように設定して動作開始(※B)
LOCAL void pwm_start_freq(INT unit, INT freq)
{
	tm_printf("pwm_start_freq: unit=%d, freq=%d\n", unit, freq);

	// 1周期のPWM出力に対するマイクロ秒単位の時間を計算(※D)
	// (＝1MHzのクロックに対するカウント値)
	// この値をカウンタ上限値に設定する
	INT cycle_us = 1000000 / freq;
	out_w(PWM(unit, COUNTERTOP), cycle_us);	// カウンタ上限値の設定(※E)
	pwm_comp0 = cycle_us / 2;		// カウンタ上限値の半分を比較値に設定(※F)

	out_w(PWM(unit, ENABLE), (1 << 0));
	out_w(PWM(unit, MODE), 0);
	out_w(PWM(unit, PRESCALER), 4);			// DIV_16でクロック1MHzを指定(※C)
	out_w(PWM(unit, LOOP), 0);
	out_w(PWM(unit, DECODER), 0);			// LOAD=commonを指定

	// 比較値を入れるRAMアドレスとサイズの設定(※G)
	out_w(PWM(unit, SEQ_PTR(0)), (UW)&pwm_comp0);
	out_w(PWM(unit, SEQ_CNT(0)), 1);
	out_w(PWM(unit, SEQ_REFRESH(0)), 0);
	out_w(PWM(unit, SEQ_ENDDELAY(0)), 0);

	// PWMのシーケンス0の動作を開始(※H)
	out_w(PWM(unit, TASKS_SEQSTART(0)), 1);
}

// unitのPWMの動作を停止(※J)
LOCAL void pwm_stop(INT unit)
{
	out_w(PWM(unit, TASKS_STOP), 1);
}

const INT speaker_pwm = 0;		// unit=0のPWMからSPEAKERに出力

// PWMを使って周波数freq(Hz)の音をptime(ミリ秒)間だけSPEAKERに出力(※K)
LOCAL void play_speaker_pwm(INT freq, INT ptime)
{
	pwm_start_freq(speaker_pwm, freq);
	tk_dly_tsk(ptime);
	pwm_stop(speaker_pwm);
}

// ドレミファソラシド(C4,D4,E4,F4,G4,A4,B4,C5)の周波数(Hz)
const INT freqs[] = { 261, 293, 329, 349, 391, 440, 493, 523 };

EXPORT void usermain(void)
{
	// SPEAKERに接続されたGPIOピン(P0.00)を出力に設定(※L)
	out_w(GPIO(P0, PIN_CNF(0)), (1 << 0));

	// PWM(unit=0)のチャンネル0をGPIOピン(P0.00)経由でSPEAKERに出力(※M)
	out_w(PWM(speaker_pwm, PSEL_OUT(0)), ((0 << 5) | (0 << 0)));

	// ドレミファ音階の再生(※N)
	for (INT octave = 0; octave < 4; octave++){
		for (INT cnt = 0; cnt < 8; cnt++){
			play_speaker_pwm((freqs[cnt] << octave), 200);
		}
		tk_dly_tsk(100);	// 1オクターブごとに100ミリ秒待つ
	}
	tk_slp_tsk(TMO_FEVR); 	// 永久待ち
}
