/*---------------------------------------------------------------
 *	PWMを使ったLED(左上のD2)の調光制御(μT-Kernel 3.0用)
 *
 *	Copyright (C) 2022-2023 by T3 WG of TRON Forum
 *---------------------------------------------------------------*/
#include <tk/tkernel.h>
#include <tm/tmonitor.h>

//-------- PWMのレジスタ定義(※A) -------------------------------
#define PWM_BASE(un)	((un) == 0 ? 0x4001c000 : (un) == 1 ? 0x40021000 : (un) == 2 ? 0x40022000 : 0x4002d000)
#define PWM(un, r)				(PWM_BASE(un) + PWM_##r)
#define PWM_TASKS_STOP			0x004
#define PWM_TASKS_SEQSTART(n)	(0x008 + (n) * 0x04)
#define PWM_LOOPSDONE			0x11C
#define PWM_ENABLE				0x500
#define PWM_MODE				0x504
#define PWM_COUNTERTOP			0x508
#define PWM_PRESCALER			0x50c
#define PWM_DECODER				0x510
#define PWM_LOOP				0x514
#define PWM_SEQ_PTR(n)			(0x520 + (n) * 0x20)
#define PWM_SEQ_CNT(n)			(0x524 + (n) * 0x20)
#define PWM_SEQ_REFRESH(n)		(0x528 + (n) * 0x20)
#define PWM_SEQ_ENDDELAY(n)		(0x52c + (n) * 0x20)
#define PWM_PSEL_OUT(n)			(0x560 + (n) * 0x04)

//-------- LED制御のためのGPIOの初期設定 ------------------------
LOCAL void led_init(void)
{
	// COL1..COL5とROW1..ROW5に接続されたGPIOピンを出力に設定
	out_w(GPIO(P0, PIN_CNF(28)), 1);	// GPIO P0.28(COL1に接続)
	out_w(GPIO(P0, PIN_CNF(11)), 1);	// GPIO P0.11(COL2に接続)
	out_w(GPIO(P0, PIN_CNF(31)), 1);	// GPIO P0.31(COL3に接続)
	out_w(GPIO(P1, PIN_CNF(05)), 1);	// GPIO P1.05(COL4に接続)
	out_w(GPIO(P0, PIN_CNF(30)), 1);	// GPIO P0.30(COL5に接続)
	out_w(GPIO(P0, PIN_CNF(21)), 1);	// GPIO P0.21(ROW1に接続)
	out_w(GPIO(P0, PIN_CNF(22)), 1);	// GPIO P0.22(ROW2に接続)
	out_w(GPIO(P0, PIN_CNF(15)), 1);	// GPIO P0.15(ROW3に接続)
	out_w(GPIO(P0, PIN_CNF(24)), 1);	// GPIO P0.24(ROW4に接続)
	out_w(GPIO(P0, PIN_CNF(19)), 1);	// GPIO P0.19(ROW5に接続)
}

//-------- GPIOの特定のピンpinの出力をval(0または1)に設定 -------
//	pinで指定された以外のピンの状態は変化しない
LOCAL void out_gpio_pin(UW port, UW pin, UW val)
{
	INT port_addr;

	if(port == 1){							// P1の場合
		if(val)
			port_addr = GPIO(P1, OUTSET);	// P1でvalが1の場合
		else							
			port_addr = GPIO(P1, OUTCLR);	// P1でvalが0の場合

	} else {								// P0の場合
		if(val)
			port_addr = GPIO(P0, OUTSET);	// P0でvalが1の場合
		else
			port_addr = GPIO(P0, OUTCLR);	// P0でvalが0の場合
	}
	out_w(port_addr, (1 << pin));			// 指定のピンに1を出力
}

// カウンタ上限値の定義
#define	TOP				1000

// 比較値COMP0を格納する16ビット幅のメモリ領域
static _UH pwm_comp0;

// パーセント単位のデューティ比dutyを指定してPWMの動作を開始(※P)
LOCAL void pwm_start_duty(INT unit, INT duty)
{
	tm_printf("pwm_start_duty: unit=%d, duty=%d%%\n", unit, duty);
	
	out_w(PWM(unit, COUNTERTOP), TOP);		// カウンタ上限値の設定

	// パーセント単位のデューティ比dutyからカウンタ比較値を計算して設定(※Q)
	pwm_comp0 = (TOP - (TOP * duty / 100));

	out_w(PWM(unit, ENABLE), (1 << 0));
	out_w(PWM(unit, MODE), 0);
	out_w(PWM(unit, PRESCALER), 4);			// DIV_16でクロック1MHzを指定(※C)
	out_w(PWM(unit, LOOP), 0);
	out_w(PWM(unit, DECODER), 0);			// LOAD=commonを指定

	// 比較値を入れるRAMアドレスとサイズの設定(※G)
	out_w(PWM(unit, SEQ_PTR(0)), (UW)&pwm_comp0);
	out_w(PWM(unit, SEQ_CNT(0)), 1);
	out_w(PWM(unit, SEQ_REFRESH(0)), 0);
	out_w(PWM(unit, SEQ_ENDDELAY(0)), 0);

	// PWMのシーケンス0の動作を開始(※H)
	out_w(PWM(unit, TASKS_SEQSTART(0)), 1);
}

// unitのPWMの動作を停止(※J)
LOCAL void pwm_stop(INT unit)
{
	out_w(PWM(unit, TASKS_STOP), 1);
}

EXPORT void usermain(void)
{
	const INT led_pwm = 0;	 	// unit=0のPWMを使ってLEDを点灯

	led_init();

	out_gpio_pin(0, 28, 0);		// GPIO P0.28-COL1に0を出力(※R)
	out_gpio_pin(0, 11, 1);		// GPIO P0.11-COL2に1を出力(※R)
	out_gpio_pin(0, 31, 1);		// GPIO P0.31-COL3に1を出力(※R)
	out_gpio_pin(1,  5, 1);		// GPIO P1.05-COL4に1を出力(※R)
	out_gpio_pin(0, 30, 1);		// GPIO P0.30-COL5に1を出力(※R)

    // PWMのチャンネル0をGPIOピン(P0.21)経由でROW1に出力(※S)
	// LED(D2)は GPIO P0.21=1(ROW1), GPIO P0.28=0(COL1) の時に点灯
	out_w(PWM(led_pwm, PSEL_OUT(0)), ((0 << 5) | (21 << 0)));

	for(INT cnt = 0; cnt < 5; cnt++){	// 以下を5回繰り返し

		// 500ミリ秒ごとにデューティ比を10%ずつ増やしながらLEDを点灯(※T)
		for(INT duty = 0; duty <= 100; duty += 10){
			pwm_start_duty(led_pwm, duty);
			tk_dly_tsk(500);
			pwm_stop(led_pwm);
		}
	}
	tk_slp_tsk(TMO_FEVR); // 永久待ち
}
